/*
 *	$Id: cpsdecode.c,v 1.4 2003/06/30 13:09:17 pwessel Exp $
 *
 *      Copyright (c) 2003 by P. Wessel
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; version 2 of the License.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *--------------------------------------------------------------------*/
/*
 *  cpsdecode - Decoding Complete PostScript (cps) files that
 *		contain embedded data and scripts/programs required
 *		to recreate the original PostScript illustration.
 *
 * Author:	Paul Wessel, SOEST, U. of Hawaii
 * Version:	2.0 [for latest release, see www.soest.hawaii.edu/pwessel/cps]
 * Date:	01-JUL-2003
 *
 * Usage:	cpsdecode PSfile [-l] [-s] [-v]
 *
 */

#include "cps.h"

main (int argc, char **argv)
{
	int i, k, n_files = 0, n, n_args;
	unsigned int n_source, n_dest;
	int verbose = FALSE, error = FALSE, listing = FALSE, read_stdin = TRUE, strip_dir = FALSE, done = FALSE;
	char buffer[BUFSIZ], tag[BUFSIZ], file[BUFSIZ], *source, *dest;
	FILE *fp = stdin, *out;

	for (i = 1; i < argc; i++) {

		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
				case '\0':
					fp = stdin;	/* For folks who are used to give - for stdin */
					break;
				case 'l':
				case 'L':
					listing = TRUE;
					break;
				case 's':
				case 'S':
					strip_dir = TRUE;
					break;
				case 'v':
				case 'V':
					verbose = TRUE;
					break;
				default:
					error = TRUE;
					break;
			}
		}
		else
			n_files++;
	}

	if (argc < 2 || error) {
		fprintf (stderr, "cpsdecode %s - Extract embedded data files from CPS PostScript files\n\n", CPS_VERSION);
		fprintf (stderr, "usage: cpsdecode [PostScriptfile(s)] [-l] [-s] [-v]\n\n");
		fprintf (stderr, "	 Give name of PostScript-file(s) or read from stdin\n\n");
		fprintf (stderr, "	 OPTIONS\n");
		fprintf (stderr, "	 -l will NOT extract files, just list the embedded files\n");
		fprintf (stderr, "	 -s will strip directory stem from names [Default uses full embedded pathname]\n");
		fprintf (stderr, "	 -v will report on progress [Default is silent]\n");
		exit (EXIT_FAILURE);
	}
	if (n_files > 0)		/* One or more files given... */
		read_stdin = FALSE;
	else				/* Read stdin instead */
		n_files = 1;
	n_args = (argc > 1) ? argc : 2;
		
	n_files = 0;

	for (k = 1; !done && k < n_args; k++) {     /* Loop over input files, if any */
	
		if (argv[k][0] == '-') continue;	/* Skip options */

		if (read_stdin) {   /* Just read standard input */
			fp = stdin;
			done = TRUE;
		}
		else if ((fp = fopen (argv[k], "r")) == NULL) {	/* Open the next PS file */
			fprintf (stderr, "cpsdecode: Could not open file %s\n", argv[k]);
			continue;
		}
		
		if (fgets (buffer, BUFSIZ, fp) == NULL) {
			fprintf (stderr, "cpsdecode: %s is an empty file\n", argv[k]);
			if (fp != stdin) fclose (fp);
			continue;
		}

		if (strncmp (buffer, "%!PS", 4)) {
			fprintf (stderr, "cpsdecode: %s is not a proper PostScript file\n", argv[k]);
			if (fp != stdin) fclose (fp);
			continue;
		}

		if (verbose && fp == stdin)
			fprintf (stderr, "cpsdecode: Reading standard input\n");
		else if (verbose)
			fprintf (stderr, "cpsdecode: Reading %s\n", argv[k]);

		while (fgets (buffer, BUFSIZ, fp)) {	/* Process one line at the time */

			if (!strncmp (buffer, "%CPS_FileBegin:", 15)) {		/* Here lies a file to decode */

				sscanf (&buffer[15], "%s %d", file, &n_dest);	/* Name of this file and full decompressed size */
			
				if (listing) {		/* Just report file name */
					fprintf (stderr, "cpsdecode: %s contains file: %s\n", argv[k], file);
				}
				else {			/* Extract the embedded data file */

					if (verbose) fprintf (stderr, "cpsdecode: Extract %s from %s\n", file, argv[k]);

					if ((n = uu_decoder (fp, &source, &n_source, tag, "%"))) {
						fprintf (stderr, "cpsdecode: Unable to allocate %d bytes - give up\n", n);
						exit (EXIT_FAILURE);
					}
					if ((dest = malloc ((size_t)n_dest)) == NULL) {
						fprintf (stderr, "cpsdecode: Unable to allocate %d bytes - give up\n", n_dest);
						exit (EXIT_FAILURE);
					}
					if (BZ2_bzBuffToBuffDecompress (dest, &n_dest, source, n_source, 0, 0) != BZ_OK) {	/* Decompress data */
						fprintf (stderr, "cpsdecode: Trouble decompressing file %s - file not extracted\n", file);
					}
					else {
						if (strip_dir) {
							for (i = strlen (file); i >= 0 && file[i] != '/'; i--);
							i++;
						}
						else
							i = 0;
						if ((out = fopen (&file[i], "wb")) == NULL) {
							fprintf (stderr, "cpsdecode: Unable to create file %s - give up\n", &file[i]);
						}
						else if (fwrite ((void *)dest, sizeof (char), (size_t)n_dest, out) != n_dest) {
							fprintf (stderr, "cpsdecode: Unable to write %d bytes to file %s - give up\n", n_dest, &file[i]);
							fclose (out);
						}
						else {	/* All went well, close up */
							fclose (out);
							n_files++;
						}
					}
					free ((void *)dest);
					free ((void *)source);
				}
			}
		}
		
		if (fp != stdin) fclose (fp);
	}

	if (!listing) {
		if (n_files == 0)
			fprintf (stderr, "cpsdecode: No files extracted. %s is probably not a CPS file\n", argv[k]);
		else if (verbose)
			fprintf (stderr, "cpsdecode: Extracted %d files\n", n_files);
	}

	exit (EXIT_SUCCESS);
}
