/*
 *	$Id: cpsencode.c,v 1.5 2003/06/30 13:09:17 pwessel Exp $
 *
 *      Copyright (c) 2003 by P. Wessel
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; version 2 of the License.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *--------------------------------------------------------------------*/
/*
 *  cpsencode - Generating Complete PostScript (cps) files that
 *		contain all the commands and the data required
 *		to recreate the plot.
 *
 * Author:	Paul Wessel, SOEST, U. of Hawaii
 * Version:	2.0 [for latest release, see www.soest.hawaii.edu/pwessel]
 * Date:	01-JUL-2003
 *
 * Usage:	cpsencode file(s) [-s] [-v] >> PSfile
 *
 */

#include "cps.h"

main (int argc, char **argv)
{
	int i, k, n_files = 0, verbose = FALSE, error = FALSE, first = TRUE, strip_dir = FALSE;
	unsigned int source_len, dest_len;
	char *source, *dest, file[128];
	FILE *fp;
	struct STAT sbuf;

	for (i = 1; i < argc; i++) {

		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
				case 's':
				case 'S':
					strip_dir = TRUE;
					break;
				case 'v':
				case 'V':
					verbose = TRUE;
					break;
				default:
					error = TRUE;
					break;
			}
		}
		else
			n_files++;
	}
	
	if (n_files == 0 || error) {
		fprintf (stderr, "cpsencode %s - Embed data files as PostScript comments\n\n", CPS_VERSION);
		fprintf (stderr, "usage: cpsencode file(s) [-s] [-v] >> psfile\n\n");
		fprintf (stderr, "	 OPTIONS\n");
		fprintf (stderr, "	 -s will strip directory stem from names [Default uses full embedded pathname]\n");
		fprintf (stderr, "	 -v will report on progress [Default is silent]\n");
		exit (EXIT_FAILURE);
	}

	for (k = 1; k < argc; k++) {

		if (argv[k][0] == '-') continue;	/* Skip options */

		if (STAT (argv[k], &sbuf)) {	/* Trouble getting stat - give up */
			fprintf (stderr, "cpsencode: Could not stat %s\n", argv[k]);
			continue;
		}
		if ((fp = fopen (argv[k], "rb")) == NULL) {
			fprintf (stderr, "cpsencode: Trouble opening file %s - file not embedded\n", argv[k]);
			continue;
		}
		source_len = sbuf.st_size;
		dest_len   = (int)(1.01 * sbuf.st_size + 600 + 0.5);	/* Worst case scenario size */
		if ((source = malloc ((size_t)source_len)) == NULL) {
			fprintf (stderr, "cpsencode: Failed to allocate %d bytes - file not embedded\n", source_len);
			continue;
		}
		if ((dest = malloc ((size_t)dest_len)) == NULL) {
			fprintf (stderr, "cpsencode: Failed to allocate %d bytes - file not embedded\n", dest_len);
			continue;
		}
		
		if (fread ((void *)source, sizeof (char), (size_t)sbuf.st_size, fp) != (size_t)sbuf.st_size) {
			fprintf (stderr, "cpsencode: Could not read %d bytes from file %s - file not embedded\n", sbuf.st_size, argv[k]);
			free ((void *)source);
			free ((void *)dest);
			fclose (fp);
			continue;
		}
		fclose (fp);
		if (BZ2_bzBuffToBuffCompress (dest, &dest_len, source, source_len, 9, 0, 30) != BZ_OK) {
			fprintf (stderr, "cpsencode: Trouble compressing file %s - file not embedded\n", argv[k]);
			free ((void *)source);
			free ((void *)dest);
			continue;
		}
		else {	/* Compress worked */
			if (first) {
				/* Logically terminate PostScript plot with EOF comment in case it has none*/

				printf ("%%%%EOF\n");

				/* Then append information about what is to come */

				printf ("%%CPS_PrivateDataBegin\n%\n");
				printf ("%%CPS_Info: ------------------------------------------------------------\n");
				printf ("%%CPS_Info: This section contains bzip2|uuencoded data files enclosed\n");
				printf ("%%CPS_Info: between pairs of CPS_FileBegin and CPS_FileEnd markers.\n");
				printf ("%%CPS_Info: Extract original data files using cpsdecode this_PS_file.\n");
				printf ("%%CPS_Info: Obtain cpsdecode from http://www.soest.hawaii.edu/pwessel/cps\n");
				printf ("%%CPS_Info: or use the following Bourne script cpsdecode.sh instead:\n");
				printf ("%%CPS_Info: Cut out, remove leading percentage comments, and chmod +x\n");
				printf ("%%CPS_Info: or extract from this file with sed -n -e 's/^%%CPS_Code: //p'\n");
				printf ("%%CPS_Code: #!/bin/sh\n");
				printf ("%%CPS_Code: if [ $# -ne 1 ]; then\n");
				printf ("%%CPS_Code: \techo \"usage: cpsdecode.sh PSfile\" >&2\n");
				printf ("%%CPS_Code: \texit\n");
				printf ("%%CPS_Code: fi\n");
				printf ("%%CPS_Code: trap \"rm -f $$; exit\" 0 1 2 15\n");
				printf ("%%CPS_Code: sed -n '/^%%CPS_PrivateDataBegin/,$p' $1 > $$\n");
				printf ("%%CPS_Code: grep -n '^%%CPS_File' $$ | tr ':' ' ' | awk '{if (NF == 5) {f = $3; s = $1} else { printf \"sed -n %%d,%%ds/^%%%%//gp %%s | uudecode\\nbzip2 -d %%s.bz2\\n\", s+1, $1-1, D, f}}' D=$$ | sh -s\n");
				printf ("%%CPS_Info: ------------------------------------------------------------\n%\n");
				first = FALSE;
			}
			if (strip_dir) {
				for (i = strlen (argv[k]); i >= 0 && argv[k][i] != '/'; i--);
				i++;
			}
			else
				i = 0;
			sprintf (file, "%s.bz2", &argv[k][i]);
			printf ("%%CPS_FileBegin: %s %d %d\n", &argv[k][i], source_len, dest_len);	/* File begin marker */
			(void) uu_encoder (stdout, dest, dest_len, file, (sbuf.st_mode & 0777), "%");
			printf ("%%CPS_FileEnd\n");							/* File end marker */
			free ((void *)source);
			free ((void *)dest);
		}
	}
	if (!first) printf ("%%CPS_PrivateDataEnd\n");

	exit (EXIT_SUCCESS);
}
