/*
 *	$Id: libuu.c,v 1.5 2003/06/30 13:09:17 pwessel Exp $
 *
 *      Copyright (c) 2003 by P. Wessel
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; version 2 of the License.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *--------------------------------------------------------------------*/
/*
 *      Include file for programs using libuu.a to do Unix-style
 *	uuencode and uudecode from within programs.  Currently
 *	the file permission mode is ignored and set to 644.
 *
 * Author:	Paul Wessel, SOEST, U. of Hawaii
 * Version:	2.0 [for latest release, see www.soest.hawaii.edu/pwessel/cps]
 * Date:	01-JUL-2003
 *
 * libuu.c: An i/o library based on uuencode and uudecode
 *
 * All that was done was to combine uuencode.c and uudecode.c
 * and replace the implicit stdin and stdout with char arrays
 * and make two functions:
 *
 * int uu_encoder (FILE *out, char *source, int n_source, char tag[], int mode, const char *prefix)
 *	out:		Pointer to stream to write encoded data
 *	source:		character array with data to be encoded
 *	n_source:	Lenght of array
 *	tag:		what uuencode calls decode-pathname: it is the
 *			name of the file that uudecode will save the
 *			decoded file to.
 *	mode:		File mode
 *	prefix:		Leading text string (if not NULL) to begin
 *			every record before the uu code is written
 *
 * int uu_decoder (FILE *in, char **dest, int *n_dest, char tag[], const char *prefix)
 *	in:		Pointer to stream from which to read data to be decoded
 *	dest:		pointer to character array to place decoded data (will be allocated)
 *	*n_dest:		Length of char array
 *	tag:		what uuencode calls decode-pathname: it is the
 *			name of the file that uudecode will save the
 *			decoded file to.
 *	prefix:		Leading text string (if not NULL) that begins
 *			every record before the uu code is read
 *
 * WIN32:  The in file pointer passed to uu_encoder and the out file pointer
 * passed to uu_decoder must refer to files opened in binary mode since we do
 * not know which files are binary or which are ascii.  We must thus deal
 * explicitly with crap like CR/LF.
  
/*------------------------------------------------------------------
 *
 * The original Copyright for uuencode and uudecode follows below.
 */
 
/* 
   Copyright (C) 1994, 1995 Free Software Foundation, Inc.

   This product is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This product is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this product; see the file COPYING.  If not, write to
   the Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.  */

/* Copyright (c) 1983 Regents of the University of California.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:
   1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
   3. All advertising materials mentioning features or use of this software
      must display the following acknowledgement:
	 This product includes software developed by the University of
	 California, Berkeley and its contributors.
   4. Neither the name of the University nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
   SUCH DAMAGE.  */

/* Reworked to GNU style by Ian Lance Taylor, ian@airs.com, August 93.  */

#include "uu.h"

/* Translation table for  standard uuencoding */
const char uu_std[64] =
{
  '`', '!', '"', '#', '$', '%', '&', '\'',
  '(', ')', '*', '+', ',', '-', '.', '/',
  '0', '1', '2', '3', '4', '5', '6', '7',
  '8', '9', ':', ';', '<', '=', '>', '?',
  '@', 'A', 'B', 'C', 'D', 'E', 'F', 'G',
  'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
  'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W',
  'X', 'Y', 'Z', '[', '\\', ']', '^', '_'
};

/* ENC is the basic 1 character encoding function to make a char printing.  */
#define ENC(Char) (uu_std[(Char) & 077])
/* Single character decode.  */
#define	DEC(Char) (((Char) - ' ') & 077)

int uu_encoder (FILE *out, char *source, unsigned int n_source, char tag[], int mode, const char *prefix)
{
	register int ch, n, close = 0, n_lines, line, k;
	register char *p;

	if (!out) out = stdout;
	
	if (prefix)
		fprintf (out, "%sbegin %o %s\n", prefix, mode, tag);
	else
		fprintf (out, "begin %o %s\n", mode, tag);

	n_lines = n_source / 45;
	for (line = k = 0; line <= n_lines; line++, k += 45) {	/* For each output line of 45 characters (last line may have less) */
		n = (line == n_lines) ? n_source - 45 * n_lines : 45;

		if (n == 0) break;

		if (prefix) fprintf (out, "%s", prefix);
		if (fputc (ENC (n), out) == EOF) break;
			
		for (p = &source[k]; n > 2; n -= 3, p += 3)
		{
			ch = *p >> 2;
			ch = ENC (ch);
			if (fputc (ch, out) == EOF) break;
			ch = ((*p << 4) & 060) | ((p[1] >> 4) & 017);
			ch = ENC (ch);
			if (fputc (ch, out) == EOF) break;
			ch = ((p[1] << 2) & 074) | ((p[2] >> 6) & 03);
			ch = ENC (ch);
			if (fputc (ch, out) == EOF) break;
			ch = p[2] & 077;
			ch = ENC (ch);
			if (fputc (ch, out) == EOF) break;
		}

		if (n != 0) break;
		if (fputc ('\n', out) == EOF) break;
	}

	while (n != 0)
	{
		char c1 = *p;
		char c2 = (n == 1) ? 0 : p[1];

		ch = c1 >> 2;
		ch = ENC (ch);
		if (fputc (ch, out) == EOF) break;

		ch = ((c1 << 4) & 060) | ((c2 >> 4) & 017);
		ch = ENC (ch);
		if (fputc (ch, out) == EOF) break;

		if (n == 1)
			ch = ENC ('\0');
		else
		{
			ch = (c2 << 2) & 074;
			ch = ENC (ch);
		}
		if (fputc (ch, out) == EOF) break;
		ch = ENC ('\0');
		if (fputc (ch, out) == EOF) break;
		fputc ('\n', out);
		n = 0;	/* old code had break; which gave warnings */
	}
	
	if (prefix) fprintf (out, "%s", prefix);
	fputc (ENC ('\0'), out);
	fputc ('\n', out);
	if (prefix) fprintf (out, "%s", prefix);
	fprintf (out, "end\n");
	
	return (0);
}

#define CHUNK 65536

int uu_decoder (FILE *in, char **dest, unsigned int *n_dest, char tag[], const char *prefix)
{
	int np, offset = 0, k = 0, n_alloc;
	char buf[2 * BUFSIZ], match[BUFSIZ], *keep;

	/* Search for header line.  */

	if (prefix) {
		offset = strlen (prefix);
		sprintf (match, "%sbegin\0", prefix);
	}
	else
		strcpy (match, "begin");
	np = 5 + offset;	/* 5 = strlen("begin") */
	
	while (1)
	{
		if (fgets (buf, sizeof (buf), in) == NULL) return 1;
		if (strncmp (buf, match, np) == 0) break;
	}

	/* Get file name */
	
	sscanf (buf, "%*s %*s %s", tag);
	
	if (prefix)
		sprintf (match, "%send\n\0", prefix);
	else
		strcpy (match, "end\n");

	n_alloc = CHUNK;
	keep = (char *) malloc ((size_t)n_alloc * sizeof (char));
	
	/* For each input line:  */
	
	while (1)
	{
		int n;
		char *p;

		if (fgets (buf, sizeof(buf), in) == NULL)
			return 3;
		if (!strcmp (buf, match))
			break;

		p = &buf[offset];

		/* N is used to avoid writing out all the characters at the end of the file.  */

		n = DEC (*p);
		for (++p; n > 0; p += 4, n -= 3)
		{
			int ch;

			if (n >= 3)
			{
				ch = DEC (p[0]) << 2 | DEC (p[1]) >> 4;
				keep[k++] = ch;
				if (k == n_alloc) {	/* Need more memory */
					n_alloc += CHUNK;
					if ((keep = (char *) realloc ((void *)keep, (size_t)n_alloc)) == NULL) return (n_alloc);
				}
				ch = DEC (p[1]) << 4 | DEC (p[2]) >> 2;
				keep[k++] = ch;
				if (k == n_alloc) {	/* Need more memory */
					n_alloc += CHUNK;
					if ((keep = (char *) realloc ((void *)keep, (size_t)n_alloc)) == NULL) return (n_alloc);
				}
				ch = DEC (p[2]) << 6 | DEC (p[3]);
				keep[k++] = ch;
				if (k == n_alloc) {	/* Need more memory */
					n_alloc += CHUNK;
					if ((keep = (char *) realloc ((void *)keep, (size_t)n_alloc)) == NULL) return (n_alloc);
				}
			}
			else
			{
				if (n >= 1)
				{
					ch = DEC (p[0]) << 2 | DEC (p[1]) >> 4;
					keep[k++] = ch;
					if (k == n_alloc) {	/* Need more memory */
						n_alloc += CHUNK;
						if ((keep = (char *) realloc ((void *)keep, (size_t)n_alloc)) == NULL) return (n_alloc);
					}
				}
				if (n >= 2)
				{
					ch = DEC (p[1]) << 4 | DEC (p[2]) >> 2;
					keep[k++] = ch;
					if (k == n_alloc) {	/* Need more memory */
						n_alloc += CHUNK;
						if ((keep = (char *) realloc ((void *)keep, (size_t)n_alloc)) == NULL) return (n_alloc);
					}
				}
			}
		}
	}

	*n_dest = k;
	if ((keep = (char *) realloc ((void *)keep, (size_t)(*n_dest) * sizeof (char))) == NULL) return (n_alloc);
	
	*dest = keep;	/* Pass char array back to calling program */
	
	return 0;
}
